#ifndef FILE_QUAKE_BSP
#define FILE_QUAKE_BSP

#include <morfbase.h>

namespace GENERAL
{
	#define FACE_POLYGON	1
	#define FACE_PATCH		2
	#define FACE_MESH		3
	#define FACE_BILLBORD	4

	#define TYPE_RAY		0
	#define TYPE_SPHERE		1
	#define TYPE_BOX		2


	struct tBSPHeader
	{
		char strID[4];				// This should always be 'IBSP'
		int version;				// This should be 0x2e for Quake 3 files
	}; 

	struct tBSPLump
	{
		int offset;					// The offset into the file for the start of this lump
		int length;					// The length in bytes for this lump
	};

	struct tBSPVertex
	{
		CVector3f vPosition;			// (x, y, z) position. 
		TMorfPointf vTextureCoord;		// (u, v) texture coordinate
		TMorfPointf vLightmapCoord;	// (u, v) lightmap coordinate
		CVector3f vNormal;				// (x, y, z) normal vector
		TColor4ub color;				// RGBA color for the vertex 

		inline tBSPVertex operator+ (const tBSPVertex &v)
		{
			tBSPVertex r;
			r.vPosition=vPosition+v.vPosition;
			r.vTextureCoord=vTextureCoord+v.vTextureCoord;
			r.vLightmapCoord=vLightmapCoord+v.vLightmapCoord;
			r.vNormal=vNormal+v.vNormal;
			r.color=color+v.color;
			return r;
		}
		inline tBSPVertex operator* (const GENERAL::SCALAR s)
		{
			tBSPVertex r;
			r.vPosition=vPosition*s;
			r.vTextureCoord=vTextureCoord*s;
			r.vLightmapCoord=vLightmapCoord*s;
			r.vNormal=vNormal*s;
			r.color=color*s;
			return r;
		}
		inline tBSPVertex &operator= (const tBSPVertex &v)
		{
			vPosition=v.vPosition;
			vTextureCoord=v.vTextureCoord;
			vLightmapCoord=v.vLightmapCoord;
			vNormal=v.vNormal;
			color=v.color;
			return *this;
		}
	};

	struct tBSPFace
	{
		int textureID;						// The index into the texture array 
		int effect;							// The index for the effects (or -1 = n/a) 
		int type;							// 1=polygon, 2=patch, 3=mesh, 4=billboard 
		int startVertIndex;					// The starting index into this face's first vertex 
				// if type = FACE_PATCH this will be filled with the Patchnumber during load after Tesselization
		int numOfVerts;						// The number of vertices for this face 
		int meshVertIndex;					// The index into the first meshvertex 
		int numMeshVerts;					// The number of mesh vertices 
		int lightmapID;						// The texture index for the lightmap 
		int lMapCorner[2];					// The face's lightmap corner in the image 
		int lMapSize[2];					// The size of the lightmap section 
		GENERAL::CVector3f lMapPos;			// The 3D origin of lightmap. 
		GENERAL::CVector3f lMapVecs[2];		// The 3D space for s and t unit vectors. 
		GENERAL::CVector3f vNormal;			// The face normal. 
		int size[2];						// The bezier patch dimensions. 
	};

	struct tBSPTexture
	{
		char strName[64];			// The name of the texture w/o the extension 
		int flags;					// The surface flags (unknown) 
		int textureType;			// The type of texture (solid, water, slime, etc..) (type & 1) = 1 (solid)
	};

	struct tBSPLightmap
	{
		GENERAL::UBYTE imageBits[128][128][3];   // The RGB data in a 128x128 image
	};

	struct tBSPNode
	{
		int plane;					// The index into the planes array 
		int front;					// The child index for the front node 
		int back;					// The child index for the back node 
		GENERAL::CVector3i min;		// The bounding box min position. 
		GENERAL::CVector3i max;				// The bounding box max position. 
	}; 

	struct tBSPLeaf
	{
		int cluster;				// The visibility cluster 
		int area;					// The area portal 
		GENERAL::CVector3i min;				// The bounding box min position 
		GENERAL::CVector3i max;				// The bounding box max position 
		int leafface;				// The first index into the face array 
		int numOfLeafFaces;			// The number of faces for this leaf 
		int leafBrush;				// The first index for into the brushes 
		int numOfLeafBrushes;		// The number of brushes for this leaf 
	}; 

	struct tBSPPlane
	{
		GENERAL::CVector3f vNormal;			// Plane normal. 
		float d;					// The plane distance from origin 
	};

	struct tBSPVisData
	{
		int numOfClusters;			// The number of clusters
		int bytesPerCluster;		// The amount of bytes (8 bits) in the cluster's bitset
		GENERAL::UBYTE *pBitsets;				// The array of bytes that holds the cluster bitsets				
	};

	struct tBSPBrush 
	{
		int brushSide;           // The starting brush side for the brush 
		int numOfBrushSides;     // Number of brush sides for the brush
		int textureID;           // The texture index for the brush
	};

	struct tBSPBrushSide 
	{
		int plane;              // The plane index
		int textureID;          // The texture index
	};

	struct  tBSPShader
	{
		char strName[64];     // The name of the shader file 
		int brushIndex;       // The brush index for this shader 
		int contentFlags;     // This is 99% of the time 5
	};



	enum eBSPLumps
	{
		kBSPEntities = 0,				// Stores player/object positions, etc...
		kBSPTextures,					// Stores texture information
		kBSPPlanes,				    // Stores the splitting planes
		kBSPNodes,						// Stores the BSP nodes
		kBSPLeafs,						// Stores the leafs of the nodes
		kBSPLeafFaces,					// Stores the leaf's indices into the faces
		kBSPLeafBrushes,				// Stores the leaf's indices into the brushes
		kBSPModels,					// Stores the info of world models
		kBSPBrushes,					// Stores the brushes info (for collision)
		kBSPBrushSides,				// Stores the brush surfaces info
		kBSPVertices,					// Stores the level vertices
		kBSPMeshIndices,				// Stores the model vertices offsets
		kBSPShaders,					// Stores the shader files (blending, anims..)
		kBSPFaces,						// Stores the faces for the level
		kBSPLightmaps,					// Stores the lightmaps for the level
		kBSPLightVolumes,				// Stores extra world lighting information
		kBSPVisData,					// Stores PVS and cluster info (visibility)
		kBSPMaxLumps					// A constant to store the number of lumps
	};

	class CQuakeBSPFile
	{
	protected:
		FILE *file;

		tBSPHeader header;
		tBSPLump lumps[GENERAL::kBSPMaxLumps];

	public:
		CQuakeBSPFile();
		bool Open(char *_fileName);

		virtual ~CQuakeBSPFile();

		char			*GetEntities(int &size);
		tBSPTexture		*GetTextures(int &size);
		tBSPPlane		*GetPlanes(int &size);
		tBSPNode		*GetNodes(int &size);
		tBSPLeaf		*GetLeafs(int &size);
		int				*GetLeafFaces(int &size);
		int				*GetLeafBrushes(int &size);
		tBSPBrush		*GetBrushes(int &size);
		tBSPBrushSide	*GetBrushSides(int &size);
		tBSPVertex		*GetVertices(int &size);
		int				*GetMeshIndices(int &size);
		tBSPShader		*GetShaders(int &size);
		tBSPFace		*GetFaces(int &size);
		tBSPLightmap	*GetLightMaps(int &size);
		void GetVisData(tBSPVisData	&visData);

		typedef std::map <std::string, std::string> TEntity;
		typedef std::multimap<std::string, TEntity> TEntities;

        void GetEntities(TEntities &_entities);



		//kBSPLightVolumes,				// Stores extra world lighting information
		//kBSPModels,					// Stores the info of world models

	};

}

#endif